import argparse
import json
import logging
import os
import sys
import time
from urllib import parse as urlparse

import requests

from Pegasus.command import Command, CompoundCommand, LoggingCommand
from Pegasus.db.ensembles import EnsembleStates, EnsembleWorkflowStates, TriggerType
from Pegasus.service.ensembles import emapp, manager
from Pegasus.service.ensembles.trigger import TriggerManager

log = logging.getLogger(__name__)

EM_PORT = os.getuid() + 7919


class EnsembleClientCommand(Command):
    def __init__(self):
        Command.__init__(self)
        self.endpoint = "http://127.0.0.1:%d/" % EM_PORT
        self.username = emapp.config["USERNAME"]
        if not self.username:
            raise Exception("Specify USERNAME in configuration")
        self.password = emapp.config["PASSWORD"]
        if not self.password:
            raise Exception("Specify PASSWORD in configuration")

    def _request(self, method, path, **kwargs):
        headers = {"accept": "application/json"}
        defaults = {"auth": (self.username, self.password), "headers": headers}
        defaults.update(kwargs)
        url = urlparse.urljoin(self.endpoint, path)
        response = requests.request(method, url, **defaults)

        if 200 <= response.status_code < 300:
            return response

        try:
            result = response.json()
            print("ERROR:", result["message"])
        except Exception:
            print("ERROR:", response.text)

        exit(1)

    def get(self, path, **kwargs):
        return self._request("get", path, **kwargs)

    def post(self, path, **kwargs):
        return self._request("post", path, **kwargs)

    def delete(self, path, **kwargs):
        return self._request("delete", path, **kwargs)

    def put(self, path, **kwargs):
        return self._request("put", path, **kwargs)

    def splitew(self, ew):
        r = ew.split(".")
        if len(r) != 2:
            self.parser.error("Invalid ENSEMBLE.WORKFLOW: %s" % ew)
        return r


class ServerCommand(LoggingCommand):
    description = "Start ensemble manager"
    usage = "%prog [options]"

    def __init__(self):
        LoggingCommand.__init__(self)
        self.parser.add_option(
            "-d",
            "--debug",
            action="store_true",
            dest="debug",
            default=None,
            help="Enable debugging",
        )

    def run(self):
        if self.options.debug:
            emapp.config.update(DEBUG=True)
            logging.getLogger().setLevel(logging.DEBUG)

        # We only start the ensemble manager if we are not debugging
        # or if we are debugging and Werkzeug is restarting. This
        # prevents us from having two ensemble managers running in
        # the debug case.
        WERKZEUG_RUN_MAIN = os.environ.get("WERKZEUG_RUN_MAIN") == "true"
        DEBUG = emapp.config.get("DEBUG", False)
        if (not DEBUG) or WERKZEUG_RUN_MAIN:
            # Make sure the environment is OK for the ensemble manager
            try:
                manager.check_environment()
            except manager.EMError as e:
                log.warning("%s: Ensemble manager disabled" % e.message)
            else:
                em_mgr = manager.EnsembleManager()
                em_mgr.start()

                trigger_mgr = TriggerManager()
                trigger_mgr.start()

        if os.getuid() == 0:
            log.fatal("The ensemble manager should not be run as root")
            exit(1)

        emapp.run(port=EM_PORT, host="127.0.0.1")

        log.info("Exiting")


def formatts(ts):
    t = time.localtime(ts)
    return time.strftime("%Y-%m-%d %H:%M:%S %Z", t)


class EnsemblesCommand(EnsembleClientCommand):
    description = "List ensembles"
    usage = "Usage: %prog ensembles"

    def run(self):
        response = self.get("/ensembles")
        result = response.json()
        fmt = "%-20s %-8s %-24s %-24s %12s %12s"
        if len(result) > 0:
            print(
                fmt
                % ("NAME", "STATE", "CREATED", "UPDATED", "MAX PLANNING", "MAX RUNNING")
            )
        for r in result:
            print(
                fmt
                % (
                    r["name"],
                    r["state"],
                    formatts(r["created"]),
                    formatts(r["updated"]),
                    r["max_planning"],
                    r["max_running"],
                )
            )


class CreateCommand(EnsembleClientCommand):
    description = "Create ensemble"
    usage = "Usage: %prog create [options] ENSEMBLE"

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser.add_option(
            "-P",
            "--max-planning",
            action="store",
            dest="max_planning",
            default=1,
            type="int",
            help="Maximum number of workflows being planned at once",
        )
        self.parser.add_option(
            "-R",
            "--max-running",
            action="store",
            dest="max_running",
            default=1,
            type="int",
            help="Maximum number of workflows running at once",
        )

    def run(self):
        if len(self.args) != 1:
            self.parser.error("Specify ENSEMBLE")

        name = self.args[0]

        request = {
            "name": name,
            "max_planning": self.options.max_planning,
            "max_running": self.options.max_running,
        }

        response = self.post("/ensembles", data=request)


def pathfind(command):
    def isexe(fn):
        return os.path.isfile(fn) and os.access(fn, os.X_OK)

    path, exe = os.path.split(command)

    if path:
        if isexe(command):
            return os.path.abspath(command)
        return None

    # Search PATH
    for path in os.environ["PATH"].split(os.pathsep):
        fn = os.path.join(path, command)
        if isexe(fn):
            return fn

    return None


class SubmitCommand(EnsembleClientCommand):
    description = "Submit workflow"
    usage = "Usage: %prog submit [options] ENSEMBLE.WORKFLOW plan_command [arg...]"

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser.disable_interspersed_args()
        self.parser.add_option(
            "-p",
            "--priority",
            action="store",
            dest="priority",
            default=0,
            help="Workflow priority",
            metavar="NUMBER",
        )

    def run(self):
        o = self.options
        p = self.parser

        if len(self.args) < 2:
            p.error("Specify ENSEMBLE.WORKFLOW and planning command")

        ew = self.args[0]
        command = self.args[1]
        args = self.args[2:]

        ensemble, workflow = self.splitew(ew)

        exe = pathfind(command)
        if exe is None:
            p.error("invalid planning command: %s" % command)

        args.insert(0, exe)

        command = '"%s"' % '" "'.join(args)

        data = {
            "name": workflow,
            "priority": o.priority,
            "basedir": os.getcwd(),
            "plan_command": command,
        }

        response = self.post("/ensembles/%s/workflows" % ensemble, data=data)


class WorkflowsCommand(EnsembleClientCommand):
    description = "List workflows in ensemble"
    usage = "Usage: %prog workflows [options] ENSEMBLE."

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser.add_option(
            "-l",
            "--long",
            action="store_true",
            dest="long",
            default=False,
            help="Show detailed output",
        )

    def run(self):
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        response = self.get("/ensembles/%s/workflows" % self.args[0])

        result = response.json()

        if len(result) == 0:
            return

        if self.options.long:
            for w in result:
                print("ID:      ", w["id"])
                print("Name:    ", w["name"])
                print("Created: ", formatts(w["created"]))
                print("Updated: ", formatts(w["updated"]))
                print("State:   ", w["state"])
                print("Priority:", w["priority"])
                print("UUID:    ", w["wf_uuid"])
                print("URL:     ", w["href"])
                print()
        else:
            fmt = "%-20s %-15s %-8s %-24s %-24s"
            print(fmt % ("NAME", "STATE", "PRIORITY", "CREATED", "UPDATED"))
            for w in result:
                print(
                    fmt
                    % (
                        w["name"],
                        w["state"],
                        w["priority"],
                        formatts(w["created"]),
                        formatts(w["updated"]),
                    )
                )


class StatusCommand(EnsembleClientCommand):
    description = "Check workflow status"
    usage = "Usage: %prog status ENSEMBLE.WORKFLOW"

    def run(self):
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE.WORKFLOW")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        ensemble, workflow = self.splitew(self.args[0])

        response = self.get("/ensembles/{}/workflows/{}".format(ensemble, workflow))

        result = response.json()

        print("ID:           %s" % result["id"])
        print("Name:         %s" % result["name"])
        print("Plan Command: %s" % result["plan_command"])
        print("Created:      %s" % formatts(result["created"]))
        print("Updated:      %s" % formatts(result["updated"]))
        print("State:        %s" % result["state"])
        print("UUID:         %s" % (result["wf_uuid"] or ""))
        print("Priority:     %s" % result["priority"])
        print("Base Dir:     %s" % result["basedir"])
        print("Submit Dir:   %s" % (result["submitdir"] or ""))
        print("Log:          %s" % result["log"])


class AnalyzeCommand(EnsembleClientCommand):
    description = "Analyze workflow status"
    usage = "Usage: %prog analyze ENSEMBLE.WORKFLOW"

    def run(self):
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE.WORKFLOW")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        ensemble, workflow = self.splitew(self.args[0])

        response = self.get(
            "/ensembles/{}/workflows/{}/analyze".format(ensemble, workflow)
        )

        sys.stdout.write(response.text)


class StateChangeCommand(EnsembleClientCommand):
    def run(self):
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        response = self.post(
            "/ensembles/%s" % self.args[0], data={"state": self.newstate}
        )
        result = response.json()

        print("State:", result["state"])


class PauseCommand(StateChangeCommand):
    description = "Pause active ensemble"
    usage = "Usage: %prog pause ENSEMBLE"
    newstate = EnsembleStates.PAUSED


class ActivateCommand(StateChangeCommand):
    description = "Activate paused or held ensemble"
    usage = "Usage: %prog activate ENSEMBLE"
    newstate = EnsembleStates.ACTIVE


class HoldCommand(StateChangeCommand):
    description = "Hold active ensemble"
    usage = "Usage: %prog hold ENSEMBLE"
    newstate = EnsembleStates.HELD


class ConfigCommand(EnsembleClientCommand):
    description = "Change ensemble configuration"
    usage = "Usage: %prog config [options] ENSEMBLE"

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser.add_option(
            "-P",
            "--max-planning",
            action="store",
            dest="max_planning",
            default=None,
            type="int",
            help="Maximum number of workflows being planned at once",
        )
        self.parser.add_option(
            "-R",
            "--max-running",
            action="store",
            dest="max_running",
            default=None,
            type="int",
            help="Maximum number of workflows running at once",
        )

    def run(self):
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        ensemble = self.args[0]

        request = {}

        if self.options.max_planning:
            request["max_planning"] = self.options.max_planning
        if self.options.max_running:
            request["max_running"] = self.options.max_running

        if len(request) == 0:
            self.parser.error("Specify --max-planning or --max-running")

        response = self.post("/ensembles/%s" % ensemble, data=request)

        result = response.json()

        print("Max Planning:", result["max_planning"])
        print("Max Running:", result["max_running"])


class WorkflowStateChangeCommand(EnsembleClientCommand):
    def run(self):
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE.WORKFLOW")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        ensemble, workflow = self.splitew(self.args[0])

        request = {"state": self.newstate}

        response = self.post(
            "/ensembles/{}/workflows/{}".format(ensemble, workflow), data=request
        )

        result = response.json()

        print("State:", result["state"])


class ReplanCommand(WorkflowStateChangeCommand):
    description = "Replan failed workflow"
    usage = "Usage: %prog replan ENSEMBLE.WORKFLOW"
    newstate = EnsembleWorkflowStates.READY


class RerunCommand(WorkflowStateChangeCommand):
    description = "Rerun failed workflow"
    usage = "Usage: %prog rerun ENSEMBLE.WORKFLOW"
    newstate = EnsembleWorkflowStates.QUEUED


class AbortCommand(WorkflowStateChangeCommand):
    description = "Abort workflow"
    usage = "Usage: %prog abort ENSEMBLE.WORKFLOW"
    newstate = EnsembleWorkflowStates.ABORTED


class PriorityCommand(EnsembleClientCommand):
    description = "Update workflow priority"
    usage = "Usage: %prog priority ENSEMBLE.WORKFLOW -p PRIORITY"

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser.add_option(
            "-p",
            "--priority",
            action="store",
            dest="priority",
            default=None,
            type="int",
            help="New workflow priority",
        )

    def run(self):
        if self.options.priority is None:
            self.parser.error("Specify -p/--priority")
        if len(self.args) == 0:
            self.parser.error("Specify ENSEMBLE.WORKFLOW")
        if len(self.args) > 1:
            self.parser.error("Invalid argument")

        ensemble, workflow = self.splitew(self.args[0])

        request = {"priority": self.options.priority}

        response = self.post(
            "/ensembles/{}/workflows/{}".format(ensemble, workflow), data=request
        )

        result = response.json()

        print("Priority:", result["priority"])


# --- Trigger Commands ---------------------------------------------------------
class CronTriggerCommand(EnsembleClientCommand):
    description = "Create a time based workflow trigger"
    usage = "Usage: pegasus-em cron-trigger ENSEMBLE TRIGGER INTERVAL WORKFLOW_SCRIPT [--timeout TIMEOUT] [--args ARG1 [ARG2 ...]]"

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser = argparse.ArgumentParser(
            usage=self.usage, description=self.description
        )

        self.parser.add_argument(
            "ensemble", type=str, help="the ensemble to which this trigger belongs"
        )

        self.parser.add_argument(
            "trigger", type=str, help="the name of the trigger to be added"
        )

        self.parser.add_argument(
            "interval",
            type=str,
            help="Duration of each trigger interval. Must be given as '<int> <s|m|h|d>' "
            "and be greater than 0 seconds",
        )

        self.parser.add_argument(
            "workflow_script", type=str, help="path to workflow script"
        )

        self.parser.add_argument(
            "-t",
            "--timeout",
            type=str,
            help="Trigger timeout. Must be given as `<int> <s|m|h|d>` and be greater than 0 seconds.",
        )

        self.parser.add_argument(
            "-a",
            "--args",
            default=None,
            type=str,
            help="CLI args to be passed to WORKFLOW_SCRIPT",
        )

    def parse(self, args):
        self.args = self.parser.parse_args(args)

    def run(self):
        request = {
            "trigger": self.args.trigger,
            "workflow_script": self.args.workflow_script,
            "workflow_args": json.dumps(self.args.args.split(" "))
            if self.args.args
            else json.dumps([]),
            "interval": self.args.interval,
            "timeout": self.args.timeout,
            "type": TriggerType.CRON.value,
        }

        response = self.post(
            "/ensembles/{e}/triggers/cron".format(e=self.args.ensemble), data=request,
        )

        print(response.json()["message"])


class FilePatternTriggerCommand(EnsembleClientCommand):
    description = "Create a file pattern based and time based workflow trigger"
    usage = "Usage: pegasus-em file-pattern-trigger ENSEMBLE TRIGGER INTERVAL WORKFLOW_SCRIPT FILE_PATTERN [FILE_PATTERN ...] [--timeout TIMEOUT] [--args ARG1 [ARG2 ...]]"

    def __init__(self):
        EnsembleClientCommand.__init__(self)
        self.parser = argparse.ArgumentParser(
            usage=self.usage, description=self.description
        )

        self.parser.add_argument(
            "ensemble", type=str, help="the ensemble to which this trigger belongs"
        )

        self.parser.add_argument(
            "trigger", type=str, help="the name of the trigger to be added"
        )

        self.parser.add_argument(
            "interval",
            type=str,
            help="Duration of each trigger interval. Must be given as '<int> <s|m|h|d>' "
            "and be greater than 0 seconds",
        )

        self.parser.add_argument(
            "workflow_script", type=str, help="path to workflow script"
        )

        self.parser.add_argument(
            "file_patterns",
            nargs="+",
            help="File pattern(s) that will be passed to glob.Glob to collect files",
        )

        self.parser.add_argument(
            "-t",
            "--timeout",
            type=str,
            help="Trigger timeout. Must be given as `<int> <s|m|h|d>` and be greater than 0 seconds.",
        )

        self.parser.add_argument(
            "-a",
            "--args",
            type=str,
            default=None,
            help="CLI args to be passed to WORKFLOW_SCRIPT",
        )

    def parse(self, args):
        self.args = self.parser.parse_args(args)

    def run(self):
        request = {
            "trigger": self.args.trigger,
            "workflow_script": self.args.workflow_script,
            "workflow_args": json.dumps(self.args.args.split(" "))
            if self.args.args
            else json.dumps([]),
            "interval": self.args.interval,
            "timeout": self.args.timeout,
            "file_patterns": json.dumps(self.args.file_patterns),
            "type": TriggerType.FILE_PATTERN.value,
        }

        response = self.post(
            "/ensembles/{e}/triggers/file_pattern".format(e=self.args.ensemble),
            data=request,
        )

        print(response.json()["message"])


# TODO: StopTriggerCommand

# TODO: ListTriggersCommand


# ------------------------------------------------------------------------------


class EnsembleCommand(CompoundCommand):
    description = "Client for ensemble management"
    commands = [
        ("server", ServerCommand),
        ("ensembles", EnsemblesCommand),
        ("create", CreateCommand),
        ("pause", PauseCommand),
        ("activate", ActivateCommand),
        ("config", ConfigCommand),
        ("submit", SubmitCommand),
        ("workflows", WorkflowsCommand),
        ("status", StatusCommand),
        ("analyze", AnalyzeCommand),
        ("replan", ReplanCommand),
        ("rerun", RerunCommand),
        ("priority", PriorityCommand),
        ("cron-trigger", CronTriggerCommand),
        ("file-pattern-trigger", FilePatternTriggerCommand),
    ]
    aliases = {
        "c": "create",
        "e": "ensembles",
        "w": "workflows",
        "sub": "submit",
        "an": "analyze",
        "st": "status",
    }


def main():
    "The entry point for pegasus-em"
    EnsembleCommand().main()



#!/usr/bin/env python3

"""
pegasus-dagman

This program is to be run as a replacement for condor_dagman inside
of a submit file. The dag can be submitted by running the command
condor_submit_dag -dagman /path/to/pegasus-dagman my.dag

Usage: pegasus-dagman [options]
"""

##
#  Copyright 2007-2010 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##
# Author : gmehta at isi dot edu
# Revision : $Revision$
__author__ = "Gaurang Mehta"
__author__ = "Mats Rynge"

import logging
import math
import os
import re
import shutil
import signal
import subprocess
import sys
import time

from Pegasus.tools import utils


def find_prog(prog, dir=[]):
    def is_prog(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(prog)
    if fpath:
        if is_prog(prog):
            return prog
    else:
        for path in dir + os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, prog)
            if is_prog(exe_file):
                return exe_file
    return None


logger = logging.getLogger("pegasus-dagman")

# Use pegasus-config to find our lib path
print("Pegasus DAGMAN is %s" % sys.argv[0])

utils.configureLogging()

SLEEP_TIME = 15
DIED_TOO_QUICKLY_TIME = 300
MONITORD_KILL_TIME = 200  # the number of seconds after pegasus-remove is called, that monitord is sent the kill signal

dagman = None

monitord = None
monitord_last_start = 0
monitord_next_start = 0
monitord_current_restarts = 0
monitord_shutdown_mode = False
monitord_shutdown_time = 0


def dagman_launch(dagman_bin, arguments=[]):
    """Launches the condor_dagman program with all
    the arguments passed to pegasus-dagman"""
    if dagman_bin is not None:

        # clean out some stuff from the env - we don't want Pegasus variables
        # from the submit host to bleed to the compute environment if the user
        # has getenv=true set
        env = os.environ.copy()
        # first saved values
        for k, v in env.copy().items():
            if re.search("^PEGASUS_ORIG_", k):
                base_k = re.sub("^PEGASUS_ORIG_", "", k)
                if env[k] == "":
                    env.pop(base_k, None)
                else:
                    env[base_k] = env[k]
        # now remove all PEGASUS_ ones
        for k, v in env.copy().items():
            if re.search("^PEGASUS_", k):
                env.pop(k, None)

        arguments.insert(0, "condor_scheduniv_exec." + os.getenv("CONDOR_ID"))
        try:
            dagman_proc = subprocess.Popen(
                arguments,
                stdout=sys.stdout,
                stderr=sys.stderr,
                executable=dagman_bin,
                env=env,
            )
            logger.info("Launched Dagman with Pid %d" % dagman_proc.pid)
        except OSError as err:
            logger.error("Could not launch Dagman.", err)
            sys.exit(1)
    else:
        logger.error("Condor Dagman not found")
        sys.exit(127)

    return dagman_proc


def monitord_launch(monitord_bin, arguments=[]):
    """Launches Monitord in condor daemon mode"""
    if monitord_bin is not None:
        try:
            # Rotate log file, if it exists
            # PM-688
            logfile = "monitord.log"
            utils.rotate_log_file(logfile)
            # we have the right name of the log file
            log = open(logfile, "a", 1)
            monitord_proc = subprocess.Popen(
                [monitord_bin, "-N", os.getenv("_CONDOR_DAGMAN_LOG")],
                stdout=log.fileno(),
                stderr=subprocess.STDOUT,
            )
            log.close()
            logger.info("Launched Monitord with Pid %d" % monitord_proc.pid)
            return monitord_proc
        except OSError as err:
            logger.error("Could not launch Monitord.", err)
    else:
        logger.error("pegausus-monitord not found")
    return None


def is_dagman_copy_to_spool():
    """Checks using condor_config_val if dagman_copy_to_spool is set
    then copy condor_dagman to the current dir "bin_dir"
    """
    condor_config_val = find_prog("condor_config_val")
    copy_to_spool = subprocess.Popen(
        [condor_config_val, "DAGMAN_COPY_TO_SPOOL"], stdout=subprocess.PIPE, shell=False
    ).communicate()[0]
    logger.info("DAGMAN_COPY_TO_SPOOL is set to %s" % copy_to_spool)
    if copy_to_spool.lower().strip() == "true":
        return True
    else:
        return False


def sighandler(signum, frame):
    """ Signal handler to catch and pass SIGTERM, SIGABRT, SIGUSR1, SIGTERM """
    #   global dagman, monitord
    logger.info("pegasus-dagman caught SIGNAL %s" % signum)
    if dagman is not None:
        os.kill(dagman.pid, signum)

    if monitord is not None:
        # PM-767 when pegasus-remove is called, internally condor_rm is called
        # that sends a SIGUSR1 to pegasus-dagman.
        # we pass that signal to condor_dagman. But for monitord, we don't, as
        # we want monitord to gracefully exit after reaching the end of dagman
        # log file, so that the stampede database records workflow failed.
        if signum == signal.SIGUSR1:
            signum = signal.SIGINT
            global monitord_shutdown_mode, monitord_shutdown_time
            monitord_shutdown_mode = True
            monitord_shutdown_time = time.time()
        else:
            # All signals other than SIGUSR1 are passed as is
            logger.info("pegasus-dagman sent signal %s to monitord" % signum)
            os.kill(monitord.pid, signum)


# -- main--------------------------------------------------------------

if __name__ == "__main__":

    # Create a new process group. PM-972: A change in HTCondor 8.2.9
    # (https://htcondor-wiki.cs.wisc.edu/index.cgi/tktview?tn=5173) means that
    # HTCondor now sets up the process group for the process, and setpgid()
    # will fail - but we can just ignore it as the process group is already set
    try:
        os.setpgid(0, 0)
    except Exception:
        pass

    signal.signal(signal.SIGTERM, sighandler)
    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGABRT, sighandler)
    signal.signal(signal.SIGUSR1, sighandler)
    signal.signal(signal.SIGUSR2, sighandler)

    copy_to_spool = is_dagman_copy_to_spool()

    # Find dagman Binary
    dagman_bin = find_prog("condor_dagman")

    if dagman_bin is not None:
        # If copy_to_spool is set copy dagman binary to dag submit directory
        if copy_to_spool:
            old_dagman_bin = dagman_bin
            dagman_bin = os.path.join(
                os.getcwd(), "condor_scheduniv_exec." + os.getenv("CONDOR_ID")
            )
            shutil.copy2(old_dagman_bin, dagman_bin)
            logger.info(
                "Copied condor_dagman from {} to {}".format(old_dagman_bin, dagman_bin)
            )

    # Launch DAGMAN
    dagman = dagman_launch(dagman_bin, sys.argv[1:])

    # Find monitord Binary
    monitord_bin = find_prog("pegasus-monitord", [os.getenv("PEGASUS_HOME") + "/bin"])

    # Launch Monitord
    monitord = monitord_launch(monitord_bin)

    dagman.poll()
    monitord.poll()

    while monitord.returncode is None or dagman.returncode is None:
        if dagman.returncode is None and monitord.returncode is not None:
            # monitord is not running
            t = time.time()
            if monitord_next_start == 0:
                logger.error("monitord is not running")
                # did the process die too quickly?
                if t - monitord_last_start < DIED_TOO_QUICKLY_TIME:
                    monitord_current_restarts += 1
                else:
                    monitord_current_restarts = 0
                # backoff with upper limit
                backoff = min(math.exp(monitord_current_restarts + 3), 3600)
                logger.info(
                    "next monitord launch scheduled in about %d seconds" % (backoff)
                )
                monitord_next_start = t + backoff - 1
            # time to restart yet?
            if monitord_next_start <= t:
                monitord_next_start = 0
                monitord_last_start = t
                monitord = monitord_launch(monitord_bin)

        # PM-767 if in shutdown mode, check to see if we need to kill monitord
        if monitord_shutdown_mode:
            t = time.time()
            if t - monitord_shutdown_time > MONITORD_KILL_TIME:
                logger.info(
                    "monitord shudown time expired. Sending SIGINT to process %d"
                    % monitord.pid
                )
                os.kill(monitord.pid, signal.SIGINT)

        # sleep in between polls
        time.sleep(SLEEP_TIME)

        monitord.poll()
        dagman.poll()

    # Dagman and Monitord have exited. Lets exit pegasus-dagman with
    # a merged returncode
    logger.info("Dagman exited with code %d" % dagman.returncode)
    logger.info("Monitord exited with code %d" % monitord.returncode)
    if copy_to_spool:
        logger.info(
            "Removing copied condor_dagman from submit directory %s" % dagman_bin
        )
        os.remove(dagman_bin)
    sys.exit(dagman.returncode & monitord.returncode)



"""
Utility functions for NetLogger modules and command-line programs
"""
__rcsid__ = "$Id: util.py 27069 2011-02-08 20:09:10Z dang $"
__author__ = "Dan Gunter (dkgunter (at) lbl.gov)"
import glob
import os
import queue
import re
import signal
import sys
import time
import traceback
from asyncore import compact_traceback
from copy import copy
from optparse import Option, OptionParser, OptionValueError, make_option

#
import Pegasus.netlogger
from Pegasus.netlogger import configobj

try:
    from hashlib import md5

    md5_new = md5
except ImportError:
    import md5

    md5_new = md5.new

## Globals

NL_HOME = "NETLOGGER_HOME"

EOF_EVENT = "netlogger.EOF"

MAGICDATE_EXAMPLES = ", ".join(
    "%s" % s
    for s in (
        "<N> weeks|days|hours|minutes|seconds time ago",
        "Today",
        "Now",
        "Tomorrow",
        "Yesterday",
        "4th [[Jan] 2003]",
        "Jan 4th 2003",
        "mm/dd/yyyy (first)",
        "dd/mm/yyyy (second)",
        "yyyy-mm-dd",
        "yyyymmdd",
        "next Tuesday",
        "last Tuesday",
    )
)

DATA_DIR = os.path.join(os.path.dirname(Pegasus.netlogger.__file__), "data")

## Exceptions


class ConfigError(Exception):
    """Use this exception for configuration-time error handling."""


class DBConnectError(Exception):
    """Used for failed database connections."""


## Classes


class ScriptOptionParser(OptionParser):
    standard_option_list = (
        make_option(
            "-q",
            "--quiet",
            action="store_true",
            dest="quiet",
            help="only report errors",
        ),
        make_option(
            "-v",
            "--verbosity",
            default=0,
            action="count",
            dest="verbosity",
            help="show verbose status information, " "repeat for even more detail",
        ),
    )


class CountingQueue(queue.Queue):
    """Wrapper around Queue that counts how many items
    were processed, for accounting purposes.
    """

    def __init__(self, type_, *args):
        queue.Queue.__init__(self, *args)
        self.n = 0
        if type_ == "get":
            self._gets, self._puts = True, False
        else:
            self._gets, self._puts = False, True

    def get(self, *args, **kwargs):
        if self._gets:
            self.n += 1
        return queue.Queue.get(self, *args, **kwargs)

    def put(self, *args, **kwargs):
        if self._puts:
            self.n += 1
        return queue.Queue.put(self, *args, **kwargs)

    def getNumProcessed(self):
        return self.n


class FIFODict:
    """A container that is limited to a fixed size.
    When that size is exceeded, every new item inserted
    results in an item removed (in FIFO order).
    Any 'old' items do not replace newer ones.
    The interface is specialized for adding an item and checking
    that it is new at the same time.
    """

    class CircularBuffer:
        """Specialized circular buffer"""

        def __init__(self, size):
            self._data = []
            self._size = size
            self._p = 0

        def getput(self, item):
            """Put item in buffer.
            If the buffer is full, return the oldest element,
            otherwise return None.
            """
            if len(self._data) == self._size:
                oldest = self._data[self._p]
                self._data[self._p] = item
                self._p = (self._p + 1) % self._size
            else:
                oldest = None
                self._data.append(item)
            return oldest

    def __init__(self, size):
        self._q = self.CircularBuffer(size)
        self._data = {}

    def add(self, key):
        """Add a key.
        If it is new, return True otherwise False.
        """
        if key in self._data:
            return False
        self._data[key] = 1
        removed = self._q.getput(key)
        if removed is not None:
            del self._data[removed]
        return True


def traceback():
    """Traceback as a string with no newlines."""
    return str(compact_traceback())


def parse_nvp(args):
    d = {}
    for arg in args:
        try:
            name, value = arg.split("=")
        except ValueError:
            pass
        d[name] = value
    return d


def tzstr():
    return "%s%02d:%02d" % (
        ("+", "-")[time.timezone > 0],
        time.timezone / 3600,
        (time.timezone - int(time.timezone / 3600) * 3600) / 60,
    )


def parseDatetime(d, utc=False):
    """Parse a datetime object, or anything that formats itself
    with isoformat(), to number of seconds since epoch.
    """
    from Pegasus.netlogger.parsers.base import parseDate

    if d is None:
        raise ValueError("date is None")
    iso = d.isoformat()
    # add 'midnight' time if none given
    if "T" not in iso:
        iso += "T00:00:00"
    # only append timezone if none is there already
    if not iso.endswith("Z") and not re.match(r".*[+-]\d\d:\d\d$", iso):
        if utc:
            iso += "Z"
        else:
            iso += tzstr()
    return parseDate(iso)


class ProgressMeter:
    """A simple textual progress meter."""

    REPORT_INTERVAL = 1000

    def __init__(self, ofile, units="lines"):
        self.ofile = ofile
        self.units = units
        self._counter = 0
        self.reset(0)

    def reset(self, n):
        self.t0 = time.time()
        self.last_report = n

    def advance(self, num=0, inc=1):
        if num == 0:
            self._counter += inc
            num = self._counter
        if num - self.last_report >= self.REPORT_INTERVAL:
            n = num - self.last_report
            dt = time.time() - self.t0
            rate = n / dt
            if rate < 1:
                dig = 2
            elif rate < 10:
                dig = 1
            else:
                dig = 0
            fmt = "[ %%5d ] %%s, rate = %%.%df %%s/sec     \r" % dig
            self.ofile.write(fmt % (num, self.units, rate, self.units))
            self.ofile.flush()
            self.reset(num)


class NullProgressMeter:
    """Substitute for ProgressMeter when you don't want anything to
    actually be printed.
    """

    def __init__(self, ofile=None, units=None):
        return

    def reset(self, n):
        pass

    def advance(self, num=0):
        pass


def mostRecentFile(dir, file_pattern, after_time=None):
    """Search 'dir' for all files matching glob 'file_pattern',
    and return the mostRecent one(s). If 'after_time' is given,
    it should be a number of seconds since the epoch UTC; no files
    will be returned if none is on or after this time.

    Returns a list of the full paths to file(s), or an empty list.
    More than one file may be returned, in the case that they have
    the same modification time.

    """
    if not os.path.isdir(dir):
        return []
    search_path = os.path.join(dir, file_pattern)
    # make a sortable list of filenames and modification times
    timed_files = [
        (os.stat_result(os.stat(fname)).st_mtime, fname)
        for fname in glob.glob(search_path)
    ]
    # if the list is empty, stop
    if not timed_files:
        return []
    # reverse sort so most-recent is first
    timed_files.sort(reverse=True)
    most_recent_time = timed_files[0][0]
    # return nothing if the most recent time is not
    # after the cutoff
    if after_time is not None and most_recent_time < int(after_time):
        return []
    # start with most recent, then append all 'ties'
    result = [timed_files[0][1]]
    i = 1
    try:
        while timed_files[i][0] == most_recent_time:
            result.append(timed_files[i][1])
            i += 1
    except IndexError:
        pass  # ran off end of list. all ties (!)
    # return all 'most recent' files
    return result


def daemonize(log=None, root_log=None, close_fds=True):
    """Make current process into a daemon."""
    # Do a double-fork so that the daemon process is completely
    # detached from its parent (it becomes a child of init).
    # For details the classic text is:
    # W. Richard Stevens, "Advanced Programming in the UNIX Environment"
    log.info("daemonize.start")
    log.debug("daemonize.fork1")
    try:
        pid = os.fork()
        if pid > 0:
            # parent: exit
            sys.exit(0)
    except OSError as err:
        log.exc("fork.1.failed", err)
        sys.exit(1)
    log.debug("daemonize.fork2")
    # child: do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # parent: exit
            sys.exit(0)
    except OSError as err:
        log.exc("daemonize.fork2.failed", err)
        sys.exit(1)
    # child: decouple from parent environment
    log.debug("daemonize.chdir_slash")
    os.chdir("/")
    try:
        os.setsid()
    except OSError:
        pass
    os.umask(0)
    if close_fds:
        # Remove log handlers that write to stdout or stderr.
        # Construct list of other log handlers' file descriptors.
        no_close = []  # list of fd's to keep open
        if root_log and len(root_log.handlers) > 0:
            console = (sys.stderr.fileno(), sys.stdout.fileno())
            for handler in root_log.handlers[:]:
                fd = handler.stream.fileno()
                # print "handler: %s, fileno=%d, console=%s" % (handler, fd, console)
                if fd in console:
                    log.removeHandler(handler)
                else:
                    no_close.append(fd)
        # Close all fd's except those that belong to non-console
        # log handlers, just discovered above.
        log.info("daemonize.close_fds", ignore=",".join(no_close))
        for fd in xrange(1024):
            if fd not in no_close:
                try:
                    os.close(fd)
                except OSError:
                    pass
    # redirect stdin, stdout, stderr to /dev/null
    log.info("daemonize.redirect")
    try:
        devnull = os.devnull
    except AttributeError:
        devnull = "/dev/null"
    os.open(devnull, os.O_RDWR)
    try:
        os.dup2(0, 1)
        os.dup2(0, 2)
    except OSError:
        pass
    log.info("daemonize.end")


def _getNumberedFiles(path):
    result = []
    for filename in glob.glob(path + ".*"):
        try:
            name, ext = filename.rsplit(".", 1)
            n = int(ext)
            result.append((n, filename))
        except (IndexError, ValueError):
            pass
    return result


def getNextNumberedFile(path, mode="w", strip=False, open_file=True):
    if strip:  # take off .<num> extension first
        path = path.rsplit(".", 1)[0]
    numbered = _getNumberedFiles(path)
    if numbered:
        numbered.sort(reverse=True)
        next_num = numbered[0][0] + 1
    else:
        next_num = 1
    next_file = "%s.%d" % (path, next_num)
    if open_file:
        return file(next_file, mode)
    else:
        return next_file


def getAllNumberedFiles(path):
    nf = _getNumberedFiles(path)
    return [x[1] for x in nf]


def getLowestNumberedFile(path, mode="r"):
    numbered = _getNumberedFiles(path)
    if numbered:
        numbered.sort()
        result = file(numbered[0][1], mode)
    else:
        result = None
    return result


class ThrottleTimer:
    """Given a ratio of time that the program should be running
    in a given time-slice, and assuming that the program is running
    continuously between calls to throttle(), periodically sleep so
    that the program is running for roughly that proportion of time.

    For example, if run_ratio is 0.1 then calling throttle() in a
    loop will cause it to sleep 90% of the time:
       tt = ThrottleTimer(0.1)
       ...
       tt.start()
       while True:
           do_something()
           tt.throttle() # sleeps here

    """

    def __init__(self, run_ratio, min_sleep_sec=0.1):
        """Create timer.

        'run_ratio' is the desired ratio of the time between calls to the
          time sleeping in ths timer.
        'min_sleep_sec' is the mininum size of the argument to time.sleep(),
          before throttle() will actually call it. This attempts to minimize
          the inaccuracy encountered with very small sleep times.
        """
        self.sleep_ratio = 1 / run_ratio - 1
        self.t0 = time.time()
        self.min_sleep_sec = min_sleep_sec

    def start(self):
        """Start the timer."""
        self.t0 = time.time()

    def throttle(self):
        """Sleep for an appropriate time.

        If that time would be less than 'min_sleep_sec' (see constructor),
        don't actually perform the sleep. Therefore, it should be safe
        to call this in a (relatively) tight loop.
        """
        t1 = time.time()
        sleep_sec = (t1 - self.t0) * self.sleep_ratio
        if sleep_sec >= self.min_sleep_sec:
            time.sleep(sleep_sec)
            self.t0 = time.time()


class NullThrottleTimer(ThrottleTimer):
    """Null class for ThrottleTimer. Cleans up calling code."""

    def __init__(self, run_ratio=None, min_sleep_sec=None):
        ThrottleTimer.__init__(self, 1)

    def start(self):
        return

    def throttle(self):
        return


class NullFile:
    """Null-object pattern for 'file' class."""

    def __init__(self, name="(null)", mode="r", buffering=None):
        self.name = name
        self.mode = mode
        self.encoding = None
        self.newlines = None
        self.softspace = 0

    def close(self):
        pass

    def closed(self):
        return True

    def fileno(self):
        return -1

    def flush(self):
        pass

    def isatty(self):
        return False

    def next(self):
        raise StopIteration()

    def read(self, n):
        return ""

    def readline(self):
        return ""

    def readlines(self):
        return []

    def seek(self, pos):
        pass

    def tell(self):
        return 0

    def write(self, data):
        return None

    def writelines(self, seq):
        return None

    def xreadlines(self):
        return self


def rm_rf(d):
    """Remove directories and their contents, recursively."""
    for path in (os.path.join(d, f) for f in os.listdir(d)):
        if os.path.isdir(path):
            rm_rf(path)
        else:
            os.unlink(path)
    os.rmdir(d)


class IncConfigObj(configobj.ConfigObj):
    """Recognize and process '#@include <file>' directives
    transparently. Do not deal with recursive references, i.e.
    ignore directives inside included files.
    """

    def __init__(self, infile, **kw):
        """Take same arguments as ConfigObj, but in the case of a file
        object or filename, process #@include statements in the file.
        """

        if not (isinstance(infile, str) or hasattr(infile, "read")):
            # not a file: stop
            configobj.ConfigObj.__init__(self, infile, **kw)
            return
        # open file
        if hasattr(infile, "read"):
            f = infile
            f.seek(0)  # rewind to start of file
        else:
            f = file(infile)
        dir_ = os.path.dirname(f.name)
        # Create list of lines that includes the included files
        lines = []
        file_lines = []  # tuple: (filename, linenum)
        i = 0
        for line in f:
            # look for include directive
            s = line.strip()
            m = re.match("@include (\"(.*)\"|'(.*)'|(\\S+))", s)
            if m:
                # This line is an @include.
                # Pick out the group that matched.
                inc_path = filter(None, m.groups()[1:])[0]
                # open the corresponding file
                if not inc_path[0] == "/":
                    inc_path = os.path.join(dir_, inc_path)
                try:
                    inc_file = file(inc_path)
                except OSError:
                    raise OSError(
                        "Cannot read file '%s' "
                        "included from '%s'" % (inc_path, f.name)
                    )
                # add contents of file to list of lines
                j = 0
                for line in inc_file:
                    j += 1
                    file_lines.append((inc_file.name, j))
                    lines.append(line)
            else:
                # This is a regular old line
                i += 1
                file_lines.append((f.name, i))
                lines.append(line)
        # Call superclass with list of lines we built
        try:
            configobj.ConfigObj.__init__(self, lines, **kw)
        except configobj.ParseError as E:
            # Report correct file and line on parse error
            m = re.search(r'line "(\d+)"', str(E))
            if m is None:
                raise
            else:
                # print file_lines
                n = int(m.group(1)) - 1
                filename, lineno = file_lines[n]
                msg = 'Invalid line {} in {}: "{}"'.format(
                    lineno, filename, lines[n].strip(),
                )
                raise configobj.ParseError(msg)

    def getHasLoggingSection(self):
        """Return True if configuration had a [logging] section,
        False otherwise.
        """
        return self._has_logging_section

    def _setHasLoggingSection(self, value):
        """Set whether configuration has a [logging] section."""
        self._has_logging_section = value


def handleSignals(*siglist):
    """Set up signal handlers.

    Input is a list of pairs of a function, and then a list of signals
    that should trigger that action, e.g.:
       handleSignals( (myfun1, (signal.SIGUSR1, signal.SIGUSR2)),
                      (myfun2, (signal.SIGTERM)) )
    """
    for action, signals in siglist:
        for signame in signals:
            if hasattr(signal, signame):
                signo = getattr(signal, signame)
                signal.signal(signo, action)


_TPAT = re.compile(r"(\d+)\s*([mhds]|minutes?|hours?|days?|seconds?)?$")
_TFAC = {
    None: 1,
    "s": 1,
    "m": 60,
    "h": 60 * 60,
    "d": 60 * 60 * 24,
    "seconds": 1,
    "minutes": 60,
    "hours": 60 * 60,
    "days": 60 * 60 * 24,
    "second": 1,
    "minute": 60,
    "hour": 60 * 60,
    "day": 60 * 60 * 24,
}


def timeToSec(s):
    """Convert time period to a number of seconds.
    Raise ValueError for invalid time, otherwise
    return a number of seconds.
    """
    s = s.lower()
    m = _TPAT.match(s)
    if m is None:
        raise ValueError("invalid time")
    g = m.groups()
    num = int(g[0])
    factor = _TFAC[g[1]]
    return num * factor


def check_timeperiod(option, opt, value):
    try:
        return timeToSec(value)
    except ValueError:
        raise OptionValueError(
            "option {}: invalid time period value: {!r}".format(opt, value)
        )


_BPAT = re.compile(r"(\d+)\s*(\S+)")
_BFAC = {
    None: 1,
    "b": 1,
    "bytes": 1,
    "kb": 1024,
    "kilobytes": 1024,
    "mb": 1024 * 1024,
    "megabytes": 1024 * 1024,
}


def sizeToBytes(s):
    """Convert a size to a number of bytes
    Return number of bytes or raise ValueError if not parseable.
    """
    m = re.match(r"(\d+)\s*(\S+)", s.lower())
    if not m:
        raise ValueError("Not of form: <num> <units>")
    value, units = m.groups()
    if units not in _BFAC:
        raise ValueError("Unrecognized units for '%s'" % s)
    return int(value) * _BFAC[units]


class ScriptOption(Option):
    TYPES = Option.TYPES + ("timeperiod",)
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["timeperiod"] = check_timeperiod


def noop(*args, **kwargs):
    """Handy no-operation function."""


def as_bool(x):
    """Convert value (possibly a string) into a boolean."""
    if x is True or x is False:
        return x
    if isinstance(x, int):
        return bool(x)
    retval = None
    if isinstance(x, str):
        retval = {
            "yes": True,
            "no": False,
            "on": True,
            "off": False,
            "1": True,
            "0": False,
            "true": True,
            "false": False,
        }.get(x.lower(), None)
    if retval is None:
        raise ValueError("Cannot convert to bool: %s" % x)
    return retval


def as_list(value, sep=" "):
    """Convert value (possibly a string) into a list.

    Raises ValueError if it's not convert-able.
    """
    retval = None
    if isinstance(value, list) or isinstance(value, tuple):
        retval = value
    elif isinstance(value, str):
        if not value:
            retval = []
        else:
            retval = value.split(sep)
    if retval is None:
        raise ValueError("Cannot convert to list: %s" % value)
    return retval


def is_stdout(fname):
    return fname == sys.stdout.name


def parseParams(opt):
    """Parse a set of name=value parameters in the input value.

    Return list of (name,value) pairs.
    Raise ValueError if a parameter is badly formatted.
    """
    params = []
    for nameval in opt:
        try:
            name, val = nameval.split("=")
        except ValueError:
            raise ValueError("Bad name=value format for '%s'" % nameval)
        params.append((name, val))
    return params


def getProgFromFile(f):
    """Get program name from __file__."""
    if f.endswith(".py"):
        f = f[:-3]
    return os.path.basename(f)


# Python 2.4-friendly uuid generator
try:
    import uuid

    def uuid1():
        return str(uuid.uuid1())


except ImportError:
    # From: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/213761
    import random
    import time

    import md5

    def uuid1():
        t = int(time.time() * 1000)
        r = int(random.random() * 100000000000000000)
        try:
            a = socket.gethostbyname(socket.gethostname())
        except Exception:
            # if we can't get a network address, just imagine one
            a = random.random() * 100000000000000000
        data = str(t) + " " + str(r) + " " + str(a)
        data = md5.md5(data).hexdigest()
        return "{}-{}-{}-{}-{}".format(
            data[0:8], data[8:12], data[12:16], data[16:20], data[20:32],
        )


"""
Word-wrap utility functions.
Examples:
   # wrap at 70 cols
   print wrap(dewrap(blob_of_text), 70)
"""


def dewrap(text):
    "Take newlines out of text and normalize whitespace."
    return " ".join(text.split())


def wrap(text, n, leader=""):
    """Word-wrap text at 'n' columns.
    The 'leader' will be inserted before each new line of text
    after the first.
    """
    if len(text) <= n:
        return text
    else:
        spc = _find_space(text, n)
        if spc < 0:
            return text
        else:
            return (
                text[:spc]
                + "\n"
                + leader
                + wrap(text[spc + 1 :].lstrip(), n, leader=leader)
            )


def _find_space(text, maxpos):
    "Find rightmost whitespace position, or -1 if none."
    p = -1
    for ws in (" ", "-"):
        p = max(p, text.rfind(ws, 0, maxpos))
    return p


def process_kvp(option, all={}, _bool={}, type="AMQP"):
    """Process a name=value option.

    Parameters:

        - option (str): "name=value" option string

    Returns: (key,value)

    Raises:  ValueError for bad format or unknown option.
    """
    parts = option.split("=", 1)
    if len(parts) != 2:
        raise ValueError("argument '%s' not in form name=value" % option)
    key, value = parts
    if all and (key not in all):
        raise ValueError("unknown {} option '{}'.".format(type, key))
    if key in _bool:
        value = as_bool(value)
    return key, value


def hash_event(e):
    """Generate and return a probabilistically unique hash code
    for event dictionary, 'e'.

    Returns: String (hexdigest) representation of the hash value.

    """
    return md5(str(e)).hexdigest()


def stringize(v):
    if isinstance(v, str):
        result = v
    elif isinstance(v, float):
        result = "%f" % v
    elif isinstance(v, int):
        result = "%d" % v
    else:
        result = str(v)
    return result



class RecordParseException(Exception):
    pass


class Token:
    START = "["
    END = "]"
    VALUE = "<value>"
    EQUALS = "="
    COMMA = ","


class RecordParser:
    def __init__(self, string):
        self.string = string
        self.index = 0
        self.next = None

    def la(self):
        if self.index >= len(self.string):
            raise RecordParseException("Unexpected end of string", self.string)
        return self.string[self.index]

    def consume(self):
        self.index += 1

    def isvalue(self, c):
        "Is it a value character?"
        if c.isspace():
            return False
        if c in '[]=,"':
            return False
        return True

    def nextToken(self):
        "Get the next token"

        # This is for the one-token lookahead
        if self.next is not None:
            tok = self.next
            self.next = None
            return tok

        while self.index < len(self.string):
            if self.la() == "[":
                self.consume()
                return (Token.START, "[")
            elif self.la() == "]":
                self.consume()
                return (Token.END, "]")
            elif self.la() == ",":
                self.consume()
                return (Token.COMMA, ",")
            elif self.la() == "=":
                self.consume()
                return (Token.EQUALS, "=")
            elif self.la().isspace():
                # Whitespace
                self.consume()
            elif self.la() == '"':
                # It is a string value
                chars = []
                self.consume()
                while self.la() != '"':
                    chars.append(self.la())
                    self.consume()
                self.consume()
                return (Token.VALUE, "".join(chars))
            else:
                # It must be a regular value
                chars = []
                while self.isvalue(self.la()):
                    chars.append(self.la())
                    self.consume()
                return (Token.VALUE, "".join(chars))

        raise RecordParseException("Unexpected end of record", self.string)

    def lt(self):
        "Look ahead one token"
        self.next = self.nextToken()
        return self.next[0]

    def expect(self, item):
        "Expect the next token to be item and return its value"
        token, value = self.nextToken()
        if token != item:
            raise RecordParseException(
                "Expected '{}', got '{}'".format(item, token), self.string
            )
        return value

    def parse(self):
        "Parse a cluster record"
        VALID_TYPES = ["cluster-summary", "seqexec-summary", "cluster-task"]

        self.expect(Token.START)
        rectype = self.expect(Token.VALUE)
        if rectype not in VALID_TYPES:
            raise RecordParseException("Invalid record type: %s" % rectype, self.string)
        record = {}

        while True:
            key = self.expect(Token.VALUE)
            self.expect(Token.EQUALS)
            value = self.expect(Token.VALUE)
            record[key] = value
            if self.lt() == Token.COMMA:
                self.expect(Token.COMMA)
            if self.lt() == Token.END:
                break
        self.expect(Token.END)

        return record



import logging
import os

import pamela
from flask import g

from Pegasus import user

log = logging.getLogger(__name__)


class BaseAuthentication:
    def __init__(self, username, password):
        self.username = username
        self.password = password

    def authenticate(self):
        raise Exception("Not implemented")

    def get_user(self):
        raise Exception("Not implemented")


class NoAuthentication(BaseAuthentication):
    def __init__(self, *args):
        # For no auth. username and password is not required
        pass

    def authenticate(self):
        # Always authenticate the user
        return True

    def get_user(self):
        # Just return info for the user running the service
        if "username" in g:
            return user.get_user_by_username(g.username)
        else:
            return user.get_user_by_uid(os.getuid())


class PAMAuthentication(BaseAuthentication):
    def authenticate(self):
        try:
            if not self.username:
                return False
            return pamela.authenticate(self.username, self.password) is None
        except Exception as e:
            log.exception(e)
            return False

    def get_user(self):
        return user.get_user_by_username(self.username)



from flask import json, make_response

JSON_HEADERS = {"Content-Type": "application/json"}


def json_api_error(e):
    response = {"message": str(e)}
    status_code = 500
    if hasattr(e, "status_code"):
        status_code = e.status_code
    return make_response(json.dumps(response), status_code, JSON_HEADERS)


def json_created(location):
    response = {"message": "created"}
    headers = {}
    headers.update(JSON_HEADERS)
    headers["Location"] = location
    return make_response(json.dumps(response), 201, headers)


def json_response(obj, status_code=200):
    return make_response(json.dumps(obj), status_code, JSON_HEADERS)



"""
properties.py

This module reads Java properties for the GriPhyN Virtual Data
System. It allows for command-line based overrides of properties
using Java's -Dk=v syntax in Python by removing initial
definitions from sys.argv during the module
initialization. Therefore, it is recommended to use this module
before parsing other command-line arguments.
"""

import logging
import os
import pwd
import re
import sys
import tempfile
import time

##
#  Copyright 2007-2010 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

# Revision : $Revision: 2012 $


# Regular expressions
re_remove_escapes = re.compile(r"\\(.)")
re_parse_property = re.compile(r"([^:= \t]+)\s*[:=]?\s*(.*)")
re_find_subs = re.compile(r"(\$\{([A-Za-z0-9._]+)\})")

re_inline_comments = re.compile("#(.*)$")  # not using it right now...

system = {}  # System properties
initial = {}  # Command-line properties

logger = logging.getLogger(__name__)

# Assemble system properties
system["file.separator"] = ","
if "JAVA_HOME" in os.environ:
    system["java.home"] = os.environ["JAVA_HOME"]
if "CLASSPATH" in os.environ:
    system["java.class.path"] = os.environ["CLASSPATH"]
if "TMP" in os.environ:
    system["java.io.tmpdir"] = os.environ["TMP"]
else:
    system["java.io.tmpdir"] = tempfile.gettempdir()
system["os.name"] = os.uname()[0]
system["os.version"] = os.uname()[2]
system["os.arch"] = os.uname()[4]
system["user.dir"] = os.getcwd()
if "HOME" in os.environ:
    system["user.home"] = os.environ["HOME"]
else:
    system["user.home"] = pwd.getpwuid(os.geteuid())[5]
if "LANG" in os.environ:
    system["user.language"] = os.environ["LANG"]
else:
    system["user.language"] = "en"
if "USER" in os.environ:
    system["user.name"] = os.environ["USER"]
elif "LOGNAME" in os.environ:
    system["user.name"] = os.environ["LOGNAME"]
else:
    system["user.name"] = pwd.getpwuid(os.geteuid())[0]
# Can be undefined
if "TZ" in os.environ:
    system["user.timezone"] = os.environ["TZ"]
# Useful, but not required
if "PEGASUS_HOME" in os.environ:
    system["pegasus.home"] = os.environ["PEGASUS_HOME"]

# Assemble command-line properties

if len(sys.argv) > 0:
    # First parameter is program name, just skip it, and remove it for now...
    program_name = sys.argv.pop(0)

    while len(sys.argv) > 0 and sys.argv[0][:2] == "-D":
        my_arg = sys.argv.pop(0)
        if my_arg == "-D":
            # k, v must be in next parameter
            if len(sys.argv) > 0:
                # Make sure we have another parameter
                my_arg = sys.argv.pop(0)
            else:
                # No, let's put the "-D" back and leave this loop
                sys.argv.insert(0, my_arg)
                break
        else:
            # remove -D from this parameter before split
            my_arg = my_arg[2:]

        try:
            k, v = my_arg.split("=", 1)
        except Exception:
            logger.warn(
                "cannot parse command-line option %s... continuing..." % (my_arg)
            )
            k = ""
        if len(k):
            if k == "pegasus.properties" or k == "pegasus.user.properties":
                logger.warn(
                    "%s is no longer supported, ignoring, please use --conf!" % (k)
                )
            else:
                logger.debug("parsed property %s..." % (my_arg))
                initial[k] = v
            # print "key:value = %s:%s" % (k, v)

    # Re-insert program_name
    sys.argv.insert(0, program_name)

# Merge the two, with command-line taking precedence over environmental variables
system.update(initial)


def parse_properties(my_file, hashref={}):
    """
    This functions parses properties from a file
    """
    # my_file is the filename or file-like of the property file to read
    # hashref contains more properties for substitution (optional)
    # global system contains yet more properties for substitution
    # returns a map of properties, possibly empty
    my_result = {}
    my_save = ""

    if isinstance(my_file, str):
        my_file = open(my_file)

    logger.debug("# parsing properties in %s..." % (my_file))

    for line in my_file:
        # line will be of type bytes if my_file was opened from a zipfile
        line = line.decode() if isinstance(line, bytes) else line
        line = line.strip(" \t")  # Remove leading and trailing spaces, tabs
        if line.startswith("!") or line.startswith("#"):
            # Skip comments
            continue
        line = line.rstrip("\n\r")  # Remove new lines, if any
        # line = re_inline_comments.sub('', line) # Remove inline comments using regular expressions
        line = line.split("#")[0]  # Remove inline comments
        line = re_remove_escapes.sub(
            r"\1", line
        )  # replace Java properties escaped special characters #!=:
        line = line.strip()  # Remove all starting and trailing whitespaces
        # Skip empty lines
        if len(line) == 0:
            continue

        if line[-1] == "\\":
            # Continuation line
            line = line[:-1]
            my_save += line
        else:
            # Regular line
            if my_save != "":
                # Append current line to previous line(s) and process
                my_save += line
                line = my_save
                my_save = ""
            logger.debug("#Property being parsed is # %s" % (line))

            # Try to parse property
            my_res = re_parse_property.search(line)
            if my_res:
                # Parse successful
                k = my_res.group(1)
                v = my_res.group(2)
                logger.debug("#Property being stored is # {} ==> {}".format(k, v))

                # Substitutions
                subs = re_find_subs.search(v)
                while subs:
                    if subs.group(1) in hashref:
                        my_newval = hashref[subs.group(1)]
                    elif subs.group(1) in system:
                        my_newval = system[subs.group(1)]
                    else:
                        my_newval = ""

                    # Make substitution
                    new_v = v[: subs.start(1)]
                    new_v += my_newval
                    new_v += v[subs.end(1) :]
                    v = new_v
                    # Search again, and loop
                    subs = re_find_subs.search(v)

                # Insert key, value into my_result
                my_result[k] = v
            else:
                logger.fatal("Illegal content in {}: {}".format(my_file, line))
                sys.exit(1)

    my_file.close()
    return my_result


class Properties:
    def __init__(self, props=None):
        # Initialize class variables
        self.m_config = props or {}

    def __delitem__(self, key):
        """
        Delete a property from the underlying dictionary.

        :param key:
        :return:
        """
        del self.m_config[key]

    def new(self, config_file=None, rundir_propfile=None):
        """
        Initialize instance variable, processing the appropriate
        properties file. config_file is the properties file passed via
        the --conf command-line option, it has the highest
        priority. rundir_propfile is the properties file in the run
        directory (specified in the braindump.txt file with the
        properties tag. It has the second highest priority. If those
        are not specified, we try to lost $(HOME)/.pegasusrc, as a
        last resort.
        """
        my_config = {}
        my_already_loaded = False

        # First, try config_file, highest priority
        if config_file is not None:
            if os.path.isfile(config_file) and os.access(config_file, os.R_OK):
                logger.debug("processing properties file %s..." % (config_file))
                my_config.update(parse_properties(config_file))
                my_already_loaded = True
            else:
                logger.warn(
                    "cannot access properties file %s... continuing..." % (config_file)
                )

        # Second, try rundir_propfile
        if not my_already_loaded and rundir_propfile is not None:
            if os.path.isfile(rundir_propfile) and os.access(rundir_propfile, os.R_OK):
                logger.debug("processing properties file %s... " % (rundir_propfile))
                my_config.update(parse_properties(rundir_propfile))
                my_already_loaded = True
            else:
                logger.warn(
                    "cannot access properties file %s... continuing..."
                    % (rundir_propfile)
                )

        # Last chance, look for $(HOME)/.pegasusrc
        if not my_already_loaded:
            if "user.home" in system:
                my_user_propfile = os.path.join(system["user.home"], ".pegasusrc")
                if os.path.isfile(my_user_propfile) and os.access(
                    my_user_propfile, os.R_OK
                ):
                    logger.debug(
                        "processing properties file %s... " % (my_user_propfile)
                    )
                    my_config.update(parse_properties(my_user_propfile))
                    my_already_loaded = True
                else:
                    # No need to complain about this
                    pass

        if not my_already_loaded:
            logger.warn("no properties file parsed whatsoever!")

        # Keep ordering of config before initial so that the -D CLI
        # properties can override any other properties
        self.m_config = my_config
        self.m_config.update(initial)

    def property(self, key, val=None):
        """
        Get and set a property
        Param: key is the property name to access
        Param: val is not None is the value to set the key
        Return: in get mode, the current value (None if not found)
        Return: in set mode, the old value (None if it didn't exist before)
        """
        my_old_val = None

        if key in self.m_config:
            my_old_val = self.m_config[key]
        if val is not None:
            self.m_config[key] = val

        return my_old_val

    def remove(self, key):
        """
        Removes a property matching the key
        :param key:
        :return: the old value if it exists
        """
        value = None
        if key in self.m_config.keys():
            value = self.m_config[key]
            del self.m_config[key]
        return value

    def keyset(self, predicate=None):
        """
        Finds a subset of keys that matches a predicate
        Param: predicate is what we match against
        Return: set of keys that match the predicate, or all keys if no predicate
        """
        if predicate is None:
            return self.m_config.keys()

        my_set = []
        for my_key in self.m_config.keys():
            if re.match(predicate, my_key):
                my_set.append(my_key)

        return my_set

    def propertyset(self, prefix, remove=False):
        """
        Finds a subset of keys that match a prefix
        Param: prefix to compare keys against
        Param: remove if True, remove prefix
        Return: Dictionary containing the matching results
        """
        my_result = {}

        for my_key in self.m_config.keys():
            # Check if it begins with prefix
            if my_key.startswith(prefix):
                if remove:
                    # Remove prefix from my_key
                    my_newkey = my_key[len(prefix) :]
                else:
                    # Keep my_key as it is
                    my_newkey = my_key
                if len(my_newkey) > 0:
                    # Only copy if my_newkey is not empty
                    my_result[my_newkey] = self.m_config[my_key]

        return my_result

    def dump(self, fn="-"):
        """
        Prints the key set in property format
        Param: fn is the name of the file to print to, defaults to stdout
        Return: number of things printed, None if error
        """
        my_count = 0

        if fn == "-":
            my_file = sys.stdout
        else:
            try:
                my_file = open(fn, "w")
            except Exception:
                logger.warn("error opening %s !" % (fn))
                return None

        # Add header
        my_file.write("# generated %s\n" % (time.asctime()))
        for my_key in self.m_config:
            # Write entry
            my_file.write("{}={}\n".format(my_key, self.m_config[my_key]))
            my_count = my_count + 1

        # Close file if not stdout
        if fn != "-":
            my_file.close()

        return my_count

    def __str__(self):
        """
        Returns string representation
        :return:
        """
        return self.m_config.__str__()


if __name__ == "__main__":
    a = Properties()
    a.new()
    print("testing finished!")



from collections import defaultdict
from io import StringIO
from typing import TextIO

from Pegasus import yaml
from Pegasus.api.errors import PegasusError
from Pegasus.api.site_catalog import (
    OS,
    Arch,
    Directory,
    FileServer,
    Grid,
    Operation,
    Scheduler,
    Site,
    SiteCatalog,
    SupportedJobs,
    _DirectoryType,
)

"""
:mod:`site_catalog` exposes an API to serialize and deserialize Pegasus's site catalog file.

Basic Usage::

    >>> from Pegasus import site_catalog
    >>> site_catalog.loads("... ")
    ...

    >>> print(site_catalog.dumps( ... ))
    ' ... '

.. moduleauthor:: Ryan Tanaka <tanaka@isi.edu>
.. moduleauthor:: Rajiv Mayani <mayani@isi.edu>
"""


__all__ = (
    "load",
    "loads",
    "dump",
    "dumps",
)


def _to_sc(d: dict) -> SiteCatalog:
    """Convert dict to SiteCatalog

    :param d: SiteCatalog represented as a dict
    :type d: dict
    :raises PegasusError: encountered error parsing
    :return: a SiteCatalog object based on d
    :rtype: SiteCatalog
    """

    try:
        sc = SiteCatalog()

        for s in d["sites"]:
            site = Site(
                s["name"],
                arch=getattr(Arch, s.get("arch").upper()) if s.get("arch") else None,
                os_type=getattr(OS, s.get("os.type").upper())
                if s.get("os.type")
                else None,
                os_release=s.get("os.release"),
                os_version=s.get("os.version"),
            )

            # add directories
            for _dir in s["directories"]:

                dir_type = None
                for enum_name, enum in _DirectoryType.__members__.items():
                    if _dir["type"] == enum.value:
                        dir_type = enum_name
                        break

                directory = Directory(
                    directory_type=getattr(Directory, dir_type),
                    path=_dir["path"],
                    shared_file_system=_dir["sharedFileSystem"],
                )

                # add file servers
                for fs in _dir["fileServers"]:
                    file_server = FileServer(
                        fs["url"], getattr(Operation, fs["operation"].upper())
                    )

                    # add profiles
                    if fs.get("profiles"):
                        file_server.profiles = defaultdict(dict, fs.get("profiles"))

                    # add file server to this directory
                    directory.add_file_servers(file_server)

                # add directory to this site
                site.add_directories(directory)

            # add grids
            if s.get("grids"):
                for gr in s.get("grids"):
                    grid = Grid(
                        getattr(Grid, gr["type"].upper()),
                        gr["contact"],
                        getattr(Scheduler, gr["scheduler"].upper()),
                        job_type=getattr(SupportedJobs, gr.get("jobtype").upper())
                        if gr.get("jobtype")
                        else None,
                    )

                    # add grid to this site
                    site.add_grids(grid)

            # add profiles
            if s.get("profiles"):
                site.profiles = defaultdict(dict, s.get("profiles"))

            # add site to sc
            sc.add_sites(site)

        return sc

    except KeyError:
        raise PegasusError("error parsing {}".format(d))


def load(fp: TextIO, *args, **kwargs) -> SiteCatalog:
    """
    Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a SiteCatalog document) to a :py:class:`~Pegasus.api.site_catalog.SiteCatalog` object.

    :param fp: file like object to load from
    :type fp: TextIO
    :return: deserialized SiteCatalog object
    :rtype: SiteCatalog
    """
    return _to_sc(yaml.load(fp))


def loads(s: str, *args, **kwargs) -> SiteCatalog:
    """
    Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance containing a SiteCatalog document) to a :py:class:`~Pegasus.api.site_catalog.SiteCatalog` object.

    :param s: string to load from
    :type s: str
    :return: deserialized SiteCatalog object
    :rtype: SiteCatalog
    """
    return _to_sc(yaml.load(s))


def dump(obj: SiteCatalog, fp: TextIO, _format="yml", *args, **kwargs) -> None:
    """
    Serialize ``obj`` as a :py:class:`~Pegasus.api.site_catalog.SiteCatalog` formatted stream to ``fp`` (a ``.write()``-supporting file-like object).

    :param obj: SiteCatalog to serialize
    :type obj: SiteCatalog
    :param fp: file like object to serialize to
    :type fp: TextIO
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :rtype: NoReturn
    """
    obj.write(fp, _format=_format)


def dumps(obj: SiteCatalog, _format="yml", *args, **kwargs) -> str:
    """
    Serialize ``obj`` to a :py:class:`~Pegasus.api.site_catalog.SiteCatalog` formatted ``str``.

    :param obj: SiteCatalog to serialize
    :type obj: SiteCatalog
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :return: SiteCatalog serialized as a string
    :rtype: str
    """
    with StringIO() as s:
        obj.write(s, _format=_format)
        s.seek(0)
        return s.read()



##
#  Copyright 2007-2011 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

"""Provides common functions used by all workflow programs."""


import calendar
import datetime
import errno
import logging
import os
import re
import shutil
import sys
import time
import traceback
import urllib
from pathlib import Path

from Pegasus import braindump
from Pegasus.vendor import attr

__all__ = ("quote", "unquote")

# Compile our regular expressions

# Used in epochdate
parse_iso8601 = re.compile(
    r"(\d{4})-?(\d{2})-?(\d{2})[ tT]?(\d{2}):?(\d{2}):?(\d{2})([.,]\d+)?([zZ]|[-+](\d{2}):?(\d{2}))"
)

# Used in out2log
re_remove_extensions = re.compile(r"(?:\.(?:rescue|dag))+$")

# Module variables
MAXLOGFILE = 1000  # For log rotation, check files from .000 to .999
jobbase = "jobstate.log"  # Default name for jobstate.log file
brainbase = "braindump.txt"  # Default name for workflow information file

logger = logging.getLogger(__name__)


class ConsoleHandler(logging.StreamHandler):
    """A handler that logs to console in the sensible way.

    StreamHandler can log to *one of* sys.stdout or sys.stderr.

    It is more sensible to log to sys.stdout by default with only error
    (logging.ERROR and above) messages going to sys.stderr. This is how
    ConsoleHandler behaves.
    """

    def __init__(self):
        logging.StreamHandler.__init__(self)
        self.stream = None  # reset it; we are not going to use it anyway

    def emit(self, record):
        if record.levelno >= logging.ERROR:
            self.__emit(record, sys.stderr)
        else:
            self.__emit(record, sys.stdout)

    def __emit(self, record, strm):
        self.stream = strm
        logging.StreamHandler.emit(self, record)

    def flush(self):
        # Workaround a bug in logging module
        # See:
        #   http://bugs.python.org/issue6333
        if self.stream and hasattr(self.stream, "flush") and not self.stream.closed:
            logging.StreamHandler.flush(self)


def configureLogging(level=logging.INFO):
    cls = logging.getLoggerClass()
    logging.TRACE = logging.DEBUG - 1

    def trace(self, message, *args, **kwargs):
        """Log a TRACE level message"""
        self.log(logging.TRACE, message, *args, **kwargs)

    logging.addLevelName(logging.TRACE, "TRACE")
    cls.trace = trace

    root = logging.getLogger()
    root.setLevel(level)
    cl = ConsoleHandler()
    formatter = logging.Formatter(
        "%(asctime)s:%(levelname)s:%(name)s(%(lineno)d): %(message)s"
    )
    cl.setFormatter(formatter)
    root.addHandler(cl)

    # logging configuration for the TriggerManager
    log_dir = Path().home() / ".pegasus/log"
    log_dir.mkdir(parents=True, exist_ok=True)

    trigger = logging.getLogger("trigger")

    # setup file handler
    log_file = str(
        log_dir
        / "{}-trigger_manager.log".format(
            datetime.datetime.now().strftime("%Y-%m-%d-T%H:%M:%S")
        )
    )
    fh = logging.FileHandler(log_file)

    # setup log format
    formatter = logging.Formatter(
        "%(asctime)s - %(name)18s - %(levelname)6s - %(message)s"
    )
    fh.setFormatter(formatter)

    trigger.addHandler(fh)


def quote(s):
    """
    Encodes a byte string using URL encoding. This replaces bytes with their
    URL-encoded equivalent %XX where XX is the hexadecimal value of the byte.
    In particular, the bytes that are encoded include:

    1. Bytes < 0x20 or >= 0x7F (C0 and C1 control codes)
    2. Quote/apostrophe (0x27), double quote (0x22), and percent (0x25)

    This will always return a byte string. If the argument is a unicode string
    then it will be utf-8 encoded before being quoted.
    """

    if not isinstance(s, bytes):
        if isinstance(s, str):
            # We need to utf-8 encode unicode strings
            s = s.encode("utf-8")
        else:
            raise TypeError("Not a string: %s" % str(s))

    buf = []
    for byte in s:
        c = chr(byte)
        if byte < 0x20 or byte >= 0x7F:
            # Any byte less than 0x20 is a control character
            # any byte >= 0x7F is either a control character
            # or a multibyte unicode character
            buf.append("%%%02X" % byte)
        elif c == "%":
            buf.append("%25")
        elif c == "'":
            buf.append("%27")
        elif c == '"':
            buf.append("%22")
        else:
            # These are regular bytes
            buf.append(c)

    return "".join(buf)


def unquote(s):
    """
    Unquote a URL-quoted string.

    This will always return a byte string with an unknown encoding. If the
    argument is a unicode string then it will be encoded in latin-1 before being
    unquoted. Latin-1 is used because it doesn't modify bytes <= 0xFF, which
    is all we expect a quoted string to contain.

    Unfortunately, this cannot return a unicode string because we have no
    way of knowing what encoding was used for the original string that was
    passed to quote().
    """
    if not isinstance(s, str):
        raise TypeError("Not a string: %s" % str(s))

    if isinstance(s, str):
        # Technically, this should not happen because
        # quote always returns a byte string, but if it was
        # passed through a database or something then it might happen
        # Latin-1 is used because it has the nice property that every
        # unicode code point <= 0xFF has the same value in latin-1
        # We ignore anything else (i.e. >0xFF) because, tecnically, it
        # should have been removed by quote()
        s = s.encode("latin-1", "ignore")

    return urllib.parse.unquote_to_bytes(s)


def isodate(now=int(time.time()), utc=False, short=False):
    """
    This function converts seconds since epoch into ISO 8601 timestamp
    """
    if utc:
        my_time_u = time.gmtime(now)
        if short:
            return time.strftime("%Y%m%dT%H%M%SZ", my_time_u)
        else:
            return time.strftime("%Y-%m-%dT%H:%M:%SZ", my_time_u)
    else:
        my_time_l = time.localtime(now)
        my_offset = int(calendar.timegm(my_time_l) - time.mktime(my_time_l))
        offset = "%+03d%02d" % (my_offset / 3600, (abs(my_offset) % 3600) / 60)
        if short:
            return time.strftime("%Y%m%dT%H%M%S", my_time_l) + offset
        else:
            return time.strftime("%Y-%m-%dT%H:%M:%S", my_time_l) + offset


def epochdate(timestamp):
    """
    This function converts an ISO timestamp into seconds since epoch
    """

    try:
        # Split date/time and timezone information
        m = parse_iso8601.search(timestamp)
        if m is None:
            logger.warn('unable to match "%s" to ISO 8601' % timestamp)
            return None
        else:
            dt = "%04d-%02d-%02d %02d:%02d:%02d" % (
                int(m.group(1)),
                int(m.group(2)),
                int(m.group(3)),
                int(m.group(4)),
                int(m.group(5)),
                int(m.group(6)),
            )
            tz = m.group(8)

        # my_time = datetime.datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")
        my_time = datetime.datetime(*(time.strptime(dt, "%Y-%m-%d %H:%M:%S")[0:6]))

        if tz.upper() != "Z":
            # no zulu time, has zone offset
            my_offset = datetime.timedelta(
                hours=int(m.group(9)), minutes=int(m.group(10))
            )

            # adjust for time zone offset
            if tz[0] == "-":
                my_time = my_time + my_offset
            else:
                my_time = my_time - my_offset

        # Turn my_time into Epoch format
        return int(calendar.timegm(my_time.timetuple()))

    except Exception:
        logger.warn('unable to parse timestamp "%s"' % timestamp)
        return None


# TODO: Remove, only used in pegasus-statistics
def create_directory(dir_name, delete_if_exists=False):
    """
    Utility method for creating directory
    @param dir_name the directory path
    @param delete_if_exists specifies whether to delete the directory if it exists
    """
    if delete_if_exists:
        if os.path.isdir(dir_name):
            logger.warning("Deleting existing directory. Deleting... " + dir_name)
            try:
                shutil.rmtree(dir_name)
            except Exception:
                logger.error("Unable to remove existing directory." + dir_name)
                sys.exit(1)
    if not os.path.isdir(dir_name):
        logger.info("Creating directory... " + dir_name)
        try:
            os.mkdir(dir_name)
        except OSError:
            logger.error("Unable to create directory." + dir_name)
            sys.exit(1)


# TODO: Remove, only used in Pegasus/monitoring/workflow.py
def find_exec(program, curdir=False, otherdirs=[]):
    """
    Determine logical location of a given binary in PATH
    """
    # program is the executable basename to look for
    # When curdir is True we also check the current directory
    # Returns fully qualified path to binary, None if not found
    my_path = os.getenv("PATH", "/bin:/usr/bin")

    for my_dir in my_path.split(":") + otherdirs:
        my_file = os.path.join(os.path.expanduser(my_dir), program)
        # Test if file is 'executable'
        if os.access(my_file, os.X_OK):
            # Found it!
            return my_file

    if curdir:
        my_file = os.path.join(os.getcwd(), program)
        # Test if file is 'executable'
        if os.access(my_file, os.X_OK):
            # Yes!
            return my_file

    # File not found
    return None


# TODO: Remove, only used in pegasus-submitdir
def write_braindump(filename, items):
    "This simply writes a dict to the file specified in braindump format"
    f = open(filename, "w")
    for k in items:
        f.write("{} {}\n".format(k, items[k]))
    f.close()


# TODO: Remove, only used in pegasus-submitdir
def read_braindump(filename):
    "This simply reads a braindump dict from the file specified"
    items = {}
    f = open(filename)
    for l in f:
        k, v = l.strip().split(" ", 1)
        k = k.strip()
        v = v.strip()
        items[k] = v
    f.close()
    return items


def _slurp_braindb(run, brain_alternate=None):
    """
    Reads extra configuration from braindump database
    Param: run is the run directory
    Returns: Dictionary with the configuration, empty if error
    """
    my_config = {}

    if brain_alternate is None:
        my_braindb = os.path.join(run, "braindump.txt")
    else:
        my_braindb = os.path.join(run, brain_alternate)

    try:
        my_file = open(my_braindb)
    except OSError:
        # Error opening file
        return my_config

    for line in my_file:
        # Remove \r and/or \n from the end of the line
        line = line.rstrip("\r\n")
        # Split the line into a key and a value
        k, v = line.split(" ", 1)

        if k == "run" and v != run and run != ".":
            logger.warn("run directory mismatch, using %s" % (run))
            my_config[k] = run
        else:
            # Remove leading and trailing whitespaces from value
            v = v.strip()
            my_config[k] = v

    # Close file
    my_file.close()

    # Done!
    logger.debug("# slurped %s" % (my_braindb))
    return my_config


def slurp_braindb(run_dir, brain_alternate=None):
    """
    Consume braindump file and return it's content as a `dict`.

    :param run_dir: [description]
    :type run_dir: [type]
    :param brain_alternate: [description], defaults to None
    :type brain_alternate: [type], optional
    :return: [description]
    :rtype: [type]
    """
    if not run_dir:
        raise ValueError("run_dir is required")

    bdump = Path(run_dir) / (
        "braindump.yml" if brain_alternate is None else brain_alternate
    )

    if bdump.exists() is False:
        # TODO: Remove this except block one backwards compatibility for
        # 4.9 is not needed.
        cfg = _slurp_braindb(run_dir, brain_alternate)
    else:
        if bdump.is_file() is False:
            raise TypeError("%s is not a valid file" % bdump.resolve())

        with bdump.open("r") as fp:
            cfg = braindump.load(fp)
            cfg = {
                k: str(v) if isinstance(v, Path) else v
                for k, v in attr.asdict(cfg).items()
            }

    return cfg


def raw_to_regular(exitcode):
    """
    This function decodes the raw exitcode into a plain format:
    For a regular exitcode, it returns a value between 0 and 127;
    For signals, it returns the negative signal number (-1 through -127)
    For failures (when exitcode < 0), it returns the special value -128
    """
    if not isinstance(exitcode, int):
        return exitcode
    if exitcode < 0:
        return -128
    if (exitcode & 127) > 0:
        # Signal
        return -(exitcode & 127)
    return exitcode >> 8


def regular_to_raw(exitcode):
    """
    This function encodes a regular exitcode into a raw exitcode.
    """
    if not isinstance(exitcode, int):
        logger.warning("exitcode not an integer!")
        return exitcode
    if exitcode == -128:
        return -1
    return exitcode << 8


def parse_exit(ec):
    """
    Parses an exit code any way possible
    Returns a string that shows what went wrong
    """
    if (ec & 127) > 0:
        my_signo = ec & 127
        my_core = ""
        if (ec & 128) == 128:
            my_core = " (core)"
        my_result = "died on signal {}{}".format(my_signo, my_core)
    elif (ec >> 8) > 0:
        my_result = "exit code %d" % (ec >> 8)
    else:
        my_result = "OK"

    return my_result


def out2log(rundir, outfile):
    """
    purpose: infer output symlink for Condor common user log
    paramtr: rundir (IN): the run directory
    paramtr: outfile (IN): the name of the out file we use
    returns: the name of the log file to use
    """

    # Get the basename
    my_base = os.path.basename(outfile)
    # NEW: Account for rescue DAGs
    my_base = my_base[: my_base.find(".dagman.out")]
    my_base = re_remove_extensions.sub("", my_base)
    # Add .log extension
    my_base = my_base + ".log"
    # Create path
    my_log = os.path.join(rundir, my_base)

    return my_log, my_base


def write_pid_file(pid_filename, ts=int(time.time())):
    """
    This function writes a pid file with name 'filename' containing
    the current pid and timestamp.
    """
    try:
        PIDFILE = open(pid_filename, "w")
        PIDFILE.write("pid %s\n" % (os.getpid()))
        PIDFILE.write("timestamp %s\n" % (isodate(ts)))
    except OSError:
        logger.error("cannot write PID file %s" % (pid_filename))
    else:
        PIDFILE.close()


def pid_running(filename):
    """
    This function takes a file containing a single line in the format
    of pid 'xxxxxx'. If the file exists, it reads the line and checks if
    the process id 'xxxxxx' is still running. The function returns True
    if the process is still running, or False if not.
    """
    # First, we check if file exists
    if os.access(filename, os.F_OK):
        try:
            # Open pid file
            PIDFILE = open(filename)

            # Look for pid line
            for line in PIDFILE:
                line = line.strip()
                if line.startswith("pid"):
                    # Get pid
                    my_pid = int(line.split(" ")[1])
                    # We are done with this file, just close it...
                    PIDFILE.close()
                    # Now let's see if process still around...
                    try:
                        os.kill(my_pid, 0)
                    except OSError as err:
                        if err.errno == errno.ESRCH:
                            # pid is not found, monitoring cannot be running
                            logger.info("pid %d not running anymore..." % (my_pid))
                            return False
                        elif err.errno == errno.EPERM:
                            # pid cannot be killed because we don't have permission
                            logger.debug(
                                "no permission to talk to pid %d..." % (my_pid)
                            )
                            return True
                        else:
                            logger.warning(
                                "unknown error while sending signal to pid %d"
                                % (my_pid)
                            )
                            logger.warning(traceback.format_exc())
                            return True
                    except Exception:
                        logger.warning(
                            "unknown error while sending signal to pid %d" % (my_pid)
                        )
                        logger.warning(traceback.format_exc())
                        return True
                    else:
                        logger.debug("pid %d still running..." % (my_pid))
                        return True

            logger.warning(
                "could not find pid line in file %s. continuing..." % (filename)
            )

            # Don't forget to close file
            PIDFILE.close()
        except Exception:
            logger.warning("error processing file %s. continuing..." % (filename))
            logger.warning(traceback.format_exc())

        return True

    # PID file doesn't exist
    return False


# TODO: Remove, only used in pegasus-statistics
def monitoring_running(run_dir):
    """
    This function takes a run directory and returns true if it appears
    that pegasus-monitord is still running, or false if it has
    finished (or perhaps it was never started).
    """
    start_file = os.path.join(run_dir, "monitord.started")
    done_file = os.path.join(run_dir, "monitord.done")

    # If monitord finished, it is not running anymore
    if os.access(done_file, os.F_OK):
        return False

    # If monitord started, it is (possibly) still running
    if os.access(start_file, os.F_OK):
        # Let's check
        return pid_running(start_file)

    # Otherwise, monitord was never executed (so it is not running right now...)
    return False


# TODO: Remove, only used in pegasus-statistics
def loading_completed(run_dir):
    """
    This function examines a run directory and returns True if all
    events were successfully processed by pegasus-monitord.
    """
    # Loading is never completed if monitoring is still running
    if monitoring_running(run_dir) is True:
        return False

    start_file = os.path.join(run_dir, "monitord.started")
    done_file = os.path.join(run_dir, "monitord.done")
    log_file = os.path.join(run_dir, "monitord.log")

    # Both started and done files need to exist...
    if (not os.access(start_file, os.F_OK)) or (not os.access(done_file, os.F_OK)):
        return False

    # Check monitord.log for loading errors...
    if os.access(log_file, os.F_OK):
        try:
            LOG = open(log_file)
            for line in LOG:
                if line.find("NL-LOAD-ERROR -->") > 0:
                    # Found loading error... event processing was not completed
                    LOG.close()
                    return False
                if line.find("KICKSTART-PARSE-ERROR -->") > 0:
                    # Found kickstart parsing error... data not fully loaded
                    LOG.close()
                    return False
                if (
                    line.find("cannot create events output... disabling event output")
                    > 0
                ):
                    # Found loader initialization error... data not loaded
                    LOG.close()
                    return False
            LOG.close()
        except OSError:
            logger.warning("could not process log file: %s" % (log_file))

    # Otherwise, return true
    return True


def rotate_log_file(source_file):
    """
    This function rotates the specified logfile.
    """

    # First we check if we have the log file
    if not os.access(source_file, os.F_OK):
        # File doesn't exist, we don't have to rotate
        return

    # Now we need to find the latest log file

    # We start from .000
    sf = 0

    while sf < MAXLOGFILE:
        dest_file = source_file + ".%03d" % (sf)
        if os.access(dest_file, os.F_OK):
            # Continue to the next one
            sf = sf + 1
        else:
            break

    # Safety check to see if we have reached the maximum number of log files
    if sf >= MAXLOGFILE:
        logger.error("%s exists, cannot rotate log file anymore!" % (dest_file))
        sys.exit(1)

    # Now that we have source_file and dest_file, try to rotate the logs
    try:
        os.rename(source_file, dest_file)
    except OSError:
        logger.error("cannot rename {} to {}".format(source_file, dest_file))
        sys.exit(1)

    # Done!
    return


def make_boolean(value):
    # purpose: convert an input string into something boolean
    # paramtr: $x (IN): a property value
    # returns: 0 (false) or 1 (true)
    my_val = str(value)
    if (
        my_val.lower() == "true"
        or my_val.lower() == "on"
        or my_val.lower() == "yes"
        or my_val.isdigit()
        and int(value) > 0
    ):
        return 1

    return 0
